from redis import Redis
from ..nuts.queue import WorkQueue

r = Redis()


def test_queue():
    r.flushall()

    wq = WorkQueue(r)

    wq.publish('TestQueuJob', {})

    assert 1 == r.scard(wq.pending_queue)



import datetime
from ..nuts.cron import Cron

cron = Cron()


def test_get_next_execution_bad_formats():
    try:
        cron.get_next_execution('* * *')
    except Exception as ex:
        assert ex.args[0] == 'Expression must have 7 postions'

    try:
        cron.get_next_execution('* * * * * * *')
    except Exception as ex:
        assert ex.args[0] == 'Cannot specify both day-of-week and day-of-month'


def test_get_next_execution():
    now = datetime.datetime(2025, 1, 9, 0, 0, 0)
    res = cron.get_next_execution('2 * * * * ? *', now)

    assert res == datetime.datetime(2025, 1, 9, 0, 0, 2, tzinfo=datetime.timezone.utc)

    now = datetime.datetime(2025, 1, 9, 0, 0, 2)
    res = cron.get_next_execution('2 * * * * ? *', now)

    assert res == datetime.datetime(2025, 1, 9, 0, 1, 2, tzinfo=datetime.timezone.utc)

    now = datetime.datetime(2025, 1, 9, 0, 0, 0)
    res = cron.get_next_execution('* * * * * ? *', now)

    assert res == datetime.datetime(2025, 1, 9, 0, 0, 1, tzinfo=datetime.timezone.utc)

    now = datetime.datetime(2025, 1, 9, 0, 0, 0)
    res = cron.get_next_execution('2 1 * * * ? *', now)

    assert res == datetime.datetime(2025, 1, 9, 0, 1, 2, tzinfo=datetime.timezone.utc)

    now = datetime.datetime(2025, 1, 9, 0, 59, 0)
    res = cron.get_next_execution('2 1 * * * ? *', now)

    assert res == datetime.datetime(2025, 1, 9, 1, 1, 2, tzinfo=datetime.timezone.utc)

    now = datetime.datetime(2025, 1, 9, 23, 0, 0)
    res = cron.get_next_execution('0 0 1 * * ? *', now)

    assert res == datetime.datetime(2025, 1, 10, 1, 0, 0, tzinfo=datetime.timezone.utc)

    now = datetime.datetime(2025, 1, 9, 23, 0, 0)
    res = cron.get_next_execution('* * 1/1 * * ? *', now)

    assert res == datetime.datetime(2025, 1, 9, 23, 0, 1, tzinfo=datetime.timezone.utc)


def test_parse_seconds():

    now = datetime.datetime(2025, 1, 9, 15, 2, 0)
    res = cron.parse_seconds('2', now)
    assert res[0] == 2
    assert res[1] is False

    now = datetime.datetime(2025, 1, 9, 15, 2, 59)
    res = cron.parse_seconds('2', now)
    assert res[0] == 2
    assert res[1] is True

    now = datetime.datetime(2025, 1, 9, 15, 2, 0)
    res = cron.parse_seconds('2/4', now)
    assert res[0] == 2
    assert res[1] is False

    now = datetime.datetime(2025, 1, 9, 15, 2, 2)
    res = cron.parse_seconds('2/4', now)
    assert res[0] == 6

    now = datetime.datetime(2025, 1, 9, 15, 2, 59)
    res = cron.parse_seconds('2/4', now)
    assert res[0] == 2
    assert res[1] is True

    now = datetime.datetime(2025, 1, 9, 15, 2, 22)
    res = cron.parse_seconds('*/4', now)
    assert res[0] == 24
    assert res[1] is False

    now = datetime.datetime(2025, 1, 9, 15, 2, 59)
    res = cron.parse_seconds('*/4', now)
    assert res[0] == 0
    assert res[1] is True


def test_parse_minutes():

    now = datetime.datetime(2025, 1, 9, 15, 0, 0)
    res = cron.parse_minutes('2', now)
    assert res[0] == 2
    assert res[1] is False

    now = datetime.datetime(2025, 1, 9, 15, 0, 0)
    res = cron.parse_minutes('2/4', now)
    assert res[0] == 2
    assert res[1] is False

    now = datetime.datetime(2025, 1, 9, 15, 2, 2)
    res = cron.parse_minutes('2/4', now)
    assert res[0] == 2

    now = datetime.datetime(2025, 1, 9, 15, 59, 59)
    res = cron.parse_minutes('2/4', now)
    assert res[0] == 2
    assert res[1] is True

    now = datetime.datetime(2025, 1, 9, 15, 22, 22)
    res = cron.parse_minutes('*/4', now)
    assert res[0] == 24
    assert res[1] is False

    now = datetime.datetime(2025, 1, 9, 15, 59, 0)
    res = cron.parse_minutes('*/4', now)
    assert res[0] == 0
    assert res[1] is True


def test_parse_hours():

    now = datetime.datetime(2025, 1, 9, 0, 0, 0)
    res = cron.parse_hours('2', now)
    assert res[0] == 2
    assert res[1] is False

    now = datetime.datetime(2025, 1, 9, 23, 0, 0)
    res = cron.parse_hours('2', now)
    assert res[0] == 2
    assert res[1] is True

    now = datetime.datetime(2025, 1, 9, 0, 0, 0)
    res = cron.parse_hours('2/4', now)
    assert res[0] == 2
    assert res[1] is False

    now = datetime.datetime(2025, 1, 9, 2, 0, 0)
    res = cron.parse_hours('2/4', now)
    assert res[0] == 2
    assert res[1] is False

    now = datetime.datetime(2025, 1, 9, 23, 0, 0)
    res = cron.parse_hours('2/4', now)
    assert res[0] == 2
    assert res[1] is True

    now = datetime.datetime(2025, 1, 9, 23, 0, 0)
    res = cron.parse_hours('*/4', now)
    assert res[0] == 0
    assert res[1] is True

    now = datetime.datetime(2025, 1, 9, 4, 0, 0)
    res = cron.parse_hours('*/4', now)
    assert res[0] == 4
    assert res[1] is False


def test_parse_day_of_month():

    now = datetime.datetime(2025, 1, 1, 0, 0, 0)
    res = cron.parse_day_of_month('2', now)
    assert res[0] == 2
    assert res[1] is False

    now = datetime.datetime(2025, 1, 1, 0, 0, 0)
    res = cron.parse_day_of_month('2/4', now)
    assert res[0] == 2
    assert res[1] is False

    now = datetime.datetime(2025, 1, 2, 0, 0, 0)
    res = cron.parse_day_of_month('2/4', now)
    assert res[0] == 2
    assert res[1] is False

    now = datetime.datetime(2025, 1, 31, 0, 0, 0)
    res = cron.parse_day_of_month('2/4', now)
    assert res[0] == 2
    assert res[1] is True

    now = datetime.datetime(2025, 1, 1, 0, 0, 0)
    res = cron.parse_day_of_month('1/1', now)
    assert res[0] == 1
    assert res[1] is False

    now = datetime.datetime(2025, 1, 1, 0, 0, 0)
    res = cron.parse_day_of_month('*/4', now)
    assert res[0] == 1
    assert res[1] is False

    now = datetime.datetime(2025, 1, 4, 0, 0, 0)
    res = cron.parse_day_of_month('*/4', now)
    assert res[0] == 5
    assert res[1] is False

    now = datetime.datetime(2025, 1, 5, 0, 0, 0)
    res = cron.parse_day_of_month('*/4', now)
    assert res[0] == 5
    assert res[1] is False


def test_parse_day_of_week():

    now = datetime.datetime(2025, 1, 1, 0, 0, 0)
    res = cron.parse_day_of_week('2', now)
    assert res[0] == 1
    assert res[1] is False

    now = datetime.datetime(2025, 1, 29, 0, 0, 0)
    res = cron.parse_day_of_week('2', now)
    assert res[0] == 29
    assert res[1] is False

    now = datetime.datetime(2025, 1, 30, 0, 0, 0)
    res = cron.parse_day_of_week('2', now)
    assert res[0] == 5
    assert res[1] is True

    now = datetime.datetime(2025, 1, 1, 0, 0, 0)
    res = cron.parse_day_of_week('2/4', now)
    assert res[0] == 1

    now = datetime.datetime(2025, 1, 7, 0, 0, 0)
    res = cron.parse_day_of_week('2/4', now)
    assert res[0] == 9

    now = datetime.datetime(2025, 1, 8, 0, 0, 0)
    res = cron.parse_day_of_week('2/4', now)
    assert res[0] == 9

    now = datetime.datetime(2025, 1, 29, 0, 0, 0)
    res = cron.parse_day_of_week('2/4', now)
    assert res[0] == 29
    assert res[1] is False

    now = datetime.datetime(2025, 1, 31, 0, 0, 0)
    res = cron.parse_day_of_week('2/5', now)
    assert res[0] == 31
    assert res[1] is False

    now = datetime.datetime(2025, 1, 1, 0, 0, 0)
    res = cron.parse_day_of_week('*', now)
    assert res[0] == 1
    assert res[1] is False


def test_parse_months():

    now = datetime.datetime(2025, 1, 1, 0, 0, 0)
    res = cron.parse_months('1', now)
    assert res[0] == 1
    assert res[1] is False

    now = datetime.datetime(2025, 12, 1, 0, 0, 0)
    res = cron.parse_months('1', now)
    assert res[0] == 1
    assert res[1] is True

    now = datetime.datetime(2025, 1, 1, 0, 0, 0)
    res = cron.parse_months('2/4', now)
    assert res[0] == 2
    assert res[1] is False


def test_parse_years():
    now = datetime.datetime(2025, 1, 1, 0, 0, 0)
    res = cron.parse_years('2025', now)
    assert res[0] == 2025
    assert res[1] is False

    now = datetime.datetime(2025, 1, 1, 0, 0, 0)
    res = cron.parse_years('2024', now)
    assert res[0] == 2025
    assert res[1] is False

    now = datetime.datetime(2025, 1, 1, 0, 0, 0)
    res = cron.parse_years('2024/2', now)
    assert res[0] == 2026
    assert res[1] is False

    now = datetime.datetime(2025, 1, 1, 0, 0, 0)
    res = cron.parse_years('*/2', now)
    assert res[0] == 2026
    assert res[1] is False



import os
import json
from redis import Redis
from .job import NutsJob
from .cron import Cron
import datetime
import logging


class Worker():
    '''
        A NUTS worker
    '''
    id: int
    redis: Redis
    scheduler: Cron
    logger: logging.Logger
    jobs: list[NutsJob]
    kwargs: dict[str, any]
    scheduled_queue: str
    pending_queue: str
    completed_queue: str
    last_run: datetime.datetime
    running_list: str

    def __init__(self, redis: Redis, jobs: list[NutsJob], **kwargs):
        self.id = os.getpid()
        self.scheduled_queue = 'nuts|jobs|scheduled'
        self.running_list = 'nuts|job|running'
        self.pending_queue = 'nuts|jobs|pending'
        self.completed_queue = 'nuts|jobs|completed'
        self.kwargs = kwargs

        self.last_run = datetime.datetime.fromtimestamp(0)

        self.redis = redis

        self.scheduler = Cron()

        self.logger = logging.getLogger(f'worker|{self.id}')

        self.jobs = []

        logging.basicConfig(level=logging.INFO)

        is_leader = self.check_leader()
        if is_leader:
            self.logger.info(f'Worker {self.id} assuming leadership')

        for job in jobs:
            j = job.Job()

            self.jobs.append(j)

            if is_leader and j.schedule:
                self.logger.info(f'Registering Cron Job {j.name}')

                next_execution = self.scheduler.get_next_execution(j.schedule).timestamp()
                scheduled_job = False
                running_job = None
                try:
                    pending = self.redis.zscan('nuts|jobs|pending', match=j.name)
                    if pending[1][0]:
                        if pending[1][1] != next_execution:
                            # Schedule has changed
                            self.redis.zrem(self.scheduled_queue, job.name)
                            self.redis.zadd(self.scheduled_queue, {job.name: next_execution})
                        else:
                            scheduled_job = True

                except Exception:
                    pass

                running_job = self.redis.hget(self.running_list, j.name)

                if not scheduled_job and not running_job:
                    self.redis.zadd(self.pending_queue, {j.name: next_execution})

    def check_leader(self) -> bool:
        is_leader = False
        leadership_check = self.redis.setnx('leader_id', self.id)

        if not leadership_check:
            leader_id = self.redis.get('leader_id')
            if leader_id == self.id:
                self.redis.expire('leader_id', 600)

                is_leader = True

        else:
            self.redis.expire('leader_id', 600)
            is_leader = True

        return is_leader

    def schedule_pending_job(self, job_name):
        self.redis.sadd(self.pending_queue, json.dumps([job_name, {}]))

    def move_scheduled_to_pending(self):
        now = round(datetime.datetime.now(datetime.timezone.utc).timestamp())
        start = round(self.last_run.timestamp())
        ready_jobs = self.redis.zrange(self.scheduled_queue, start=start, end=now, withscores=True)

        if len(ready_jobs) > 0:
            latest = max(j[0] for j in ready_jobs)
            self.redis.zremrangebyscore(self.scheduled_queue, min=self.last_run.timestamp(), max=latest)

        for job in ready_jobs:
            self.schedule_pending_job(job[0])

    def move_pending_to_running(self, job: NutsJob, job_args: list):
        now = datetime.datetime.now(datetime.timezone.utc).isoformat()
        self.redis.hset(self.running_list, f'{self.id}|{job.name}', json.dumps({'timestamp': now, 'args': job_args}))

    def remove_running(self, job: NutsJob):

        self.redis.hdel(self.running_list, f'{self.id}|{job.name}')

    def move_to_completed(self, job: NutsJob):
        self.redis.sadd(self.completed_queue, job.name)

    def queue_completed_jobs(self):

        completed_jobs = self.redis.smembers(self.completed_queue)

        for job_name in completed_jobs:
            self.redis.srem(self.completed_queue, job_name)
            job = [j for j in self.jobs if j.name == job_name.decode()][0]

            next_execution = self.scheduler.get_next_execution(job.schedule).timestamp()

            self.redis.zadd(self.scheduled_queue, {job.name: next_execution})

    def run(self):

        is_leader = self.check_leader()
        if is_leader:
            # Move ready jobs to the pending queue
            self.move_scheduled_to_pending()

        try:
            data = self.redis.spop(self.pending_queue, 1)
            [job_name, job_args] = json.loads(data[0])

            job = [j for j in self.jobs if j.name == job_name][0]

            self.move_pending_to_running(job, job_args)

            result = job.run(**job_args, **self.kwargs)

            self.remove_running(job)

            if job.success:
                loggable_result = job.result if job.result and job.result is not None else result
                msg = f'SUCCESS: {job.name}, {loggable_result}'

                self.logger.info(msg)
            else:
                self.logger.error(f'Error running job {job.name}: {job.error}')
            if job.schedule:
                self.move_job_to_completed(job)

        except Exception as ex:
            self.logger.error(ex)

        # Post Execution

        if is_leader:
            self.queue_completed_jobs()




import json
from redis import Redis
import logging


class WorkQueue():
    redis: Redis
    logger: logging.Logger

    pending_queue: str

    def __init__(self, redis: Redis, **kwargs):
        self.redis = redis
        self.logger = logging.getLogger(kwargs.get('logger', 'nuts|workqueue'))
        self.pending_queue = 'nuts|jobs|pending'

    def publish(self, job_name: str, job_parameters: object):
        self.redis.sadd(self.pending_queue, json.dumps([job_name, job_parameters]))



from typing import Union, Any


class NutsJob():
    '''
        A Nuts Job
    '''
    name: str
    schedule: Union[str, None]
    success: bool
    result: Any
    error: Exception

    def __init__(self):
        self.name = None
        self.schedule = None
        self.success = False
        self.result = None
        self.error = None



import datetime
from dateutil.relativedelta import relativedelta
import calendar


class Cron():

    def __get_start_frequency(self, input: str, base: str = None) -> list[str]:
        try:
            [start, frequency] = input.split('/')
            if start == '*':
                start = base if base is not None else '0'
            return [int(start), int(frequency)]

        except Exception:
            raise Exception(f'Invalid expression {input}: Valid examples include: */int, int/int')

    def __get_week_structure(self, year: int, month: int) -> list[int]:
        start_of_month = calendar.monthrange(year, month)[0].value
        week_structure = []
        while len(week_structure) < 7:
            if start_of_month < 8:
                week_structure.append(start_of_month)
                start_of_month += 1
            else:
                start_of_month = 1
        return week_structure

    def get_next_execution(self, schedule: str, now: datetime.datetime = datetime.datetime.now(datetime.timezone.utc)) -> int:
        try:
            [seconds, minutes, hours, day_of_week, month, day_of_month, year] = schedule.split(' ')
        except Exception:
            raise Exception('Expression must have 7 postions')

        if '?' not in [day_of_week, day_of_month]:
            raise Exception('Cannot specify both day-of-week and day-of-month')

        [next_seconds, inc_minute] = self.parse_seconds(seconds, now)
        if inc_minute:
            now = now + datetime.timedelta(minutes=1)
        [next_minutes, inc_hour] = self.parse_minutes(minutes, now)
        if inc_hour:
            now = now + datetime.timedelta(hours=1)
        [next_hours, inc_day] = self.parse_hours(hours, now)
        if inc_day:
            now = now + datetime.timedelta(days=1)
        if day_of_week == '?':
            [next_days, inc_month] = self.parse_day_of_month(day_of_month, now)

        else:
            [next_days, inc_month] = self.parse_day_of_week(day_of_week, now)

        if inc_month:
            now = now + relativedelta(months=1)
        [next_months, inc_year] = self.parse_months(month, now)
        if inc_year:
            now = now + relativedelta(years=1)

        [next_year, _] = self.parse_years(year, now)
        res = datetime.datetime(next_year, next_months, next_days, next_hours, next_minutes, next_seconds, tzinfo=datetime.timezone.utc)

        return res

    def parse_seconds(self, seconds: str, now: datetime.datetime) -> list[int, bool]:
        if seconds == '*':
            seconds = '0/1'

        if '/' not in seconds:
            start = int(seconds)
            frequency = 60
        else:
            [start, frequency] = self.__get_start_frequency(seconds)

        upcoming = [i for i in range(start, 60, frequency) if i > now.second]
        try:
            return [upcoming[0], False]
        except Exception:
            return [start, True]

    def parse_minutes(self, minutes: str, now: datetime.datetime) -> list[int, False]:
        if minutes == '*':
            minutes = '0/1'

        if '/' not in minutes:
            start = int(minutes)
            frequency = 60
        else:
            [start, frequency] = self.__get_start_frequency(minutes)

        upcoming = [i for i in range(start, 60, frequency) if i >= now.minute]
        try:
            return [upcoming[0], False]
        except Exception:
            return [start, True]

    def parse_hours(self, hours: str, now: datetime.datetime) -> list[int, bool]:
        if hours == '*':
            hours = '0/1'

        if '/' not in hours:
            start = int(hours)
            frequency = 24
        else:
            [start, frequency] = self.__get_start_frequency(hours)

        upcoming = [i for i in range(start, 24, frequency) if i >= now.hour]
        try:
            return [upcoming[0], False]
        except Exception:
            return [start, True]

    def parse_day_of_month(self, days: str, now: datetime.datetime) -> list[int, bool]:
        end_of_month = calendar.monthrange(now.year, now.month)[1]
        if days == '*':
            days = '1/1'
        if '/' not in days:
            start = int(days)
            frequency = end_of_month
        else:
            [start, frequency] = self.__get_start_frequency(days, '1')

        upcoming = [i for i in range(start, end_of_month + 1, frequency) if i >= now.day]
        try:
            return [upcoming[0], False]
        except Exception:
            return [start, True]

    def __get_upcoming_weekdays(self, days: str, week_structure: list[int], now: datetime.datetime, frequency: int = None) -> list[int]:
        end_of_month = calendar.monthrange(now.year, now.month)[1] + 1
        month_days = list(range(1, end_of_month))
        weeks = [month_days[i * 7:(i+1) * 7] for i in range((len(month_days) + 7 - 1) // 7)]
        dow_idx = week_structure.index(int(days))

        upcoming = []

        if frequency:
            start = weeks[0][dow_idx]
            upcoming = [i for i in range(start, end_of_month, frequency) if i >= now.day]
        else:
            for w in weeks:
                try:
                    d = w[dow_idx]
                    if d >= now.day:
                        upcoming.append(d)
                except Exception:
                    pass
        return upcoming

    def parse_day_of_week(self, days: str, now: datetime.datetime) -> int:
        week_structure = self.__get_week_structure(now.year, now.month)
        if days == '*':
            days = f'{week_structure[0]}/1'

        if '/' not in days:
            start = int(days)

            upcoming = self.__get_upcoming_weekdays(start, week_structure, now)
            try:
                return [upcoming[0], False]
            except Exception:
                # Need to shift forward to the next month
                next_month = now + datetime.timedelta(days=calendar.monthrange(now.year, now.month)[1] - now.day + 1)
                week_structure = self.__get_week_structure(next_month.year, next_month.month)
                upcoming = self.__get_upcoming_weekdays(start, week_structure, next_month)

                return [upcoming[0], True]

        else:
            [start, frequency] = self.__get_start_frequency(days, '1')
            upcoming = self.__get_upcoming_weekdays(start, week_structure, now, frequency)
            try:
                return [upcoming[0], False]
            except Exception:
                # Need to shift forward to the next month
                now = now + datetime.timedelta(days=calendar.monthrange(now.year, now.month)[1] - now.day + 1)
                week_structure = self.__get_week_structure(now.year, now.month)
                upcoming = self.__get_upcoming_weekdays(start, week_structure, now, frequency)

                return [upcoming[0], True]

    def parse_months(self, months: str, now: datetime.datetime) -> int:
        if months == '*':
            months = '1/1'

        if '/' not in months:
            start = int(months)
            frequency = 12
        else:
            [start, frequency] = self.__get_start_frequency(months, '1')

        upcoming = [i for i in range(start, 13, frequency) if i >= now.month]
        try:
            return [upcoming[0], False]
        except Exception:
            return [start, True]

    def parse_years(self, years: str, now: datetime.datetime) -> int:
        if years == '*':
            years = '1970/1'

        if '/' not in years:
            start = int(years)
            frequency = 1
        else:
            [start, frequency] = self.__get_start_frequency(years, '1970')

        upcoming = [i for i in range(start, now.year + frequency, frequency) if i >= now.year]
        try:
            return [upcoming[0], False]
        except Exception:
            return [start, True]


if __name__ == '__main__':
    cron = Cron()
    cron.get_next_execution('2/4 * * * * ? *')



from ..nuts.job import NutsJob


def test_job():
    class Job(NutsJob):
        def __init__(self):
            self.name = 'job'

        def run(self, args):
            self.result = args.get('base') + 1
            self.success = True

    job = Job()
    job.run({'base': 1})

    assert job.name == 'job'
    assert job.success is True
    assert job.result == 2


def test_job_failure():
    class FailJob(NutsJob):
        def __init__(self):
            self.name = 'fail_job'

        def run(self, args):
            self.success = False




from ..nuts.worker import Worker
from redis import Redis
from .fixtures.jobs import add_one, scheduled_job
import json

jobs = [add_one, scheduled_job]
r = Redis()


def middleware_func():
    return 'I am middleware'


worker = Worker(redis=r, jobs=jobs, middleware=middleware_func)

r.flushall()


def test_worker():
    r.sadd(worker.pending_queue, json.dumps(['AddOne', {'base': 1}]))

    worker.run()

    assert 1 == 1


def test_scheduled_job():
    r.sadd(worker.completed_queue, 'ScheduledJob')

    worker.queue_completed_jobs()

    scheduled_jobs = r.zscan(worker.scheduled_queue, match='ScheduledJob')
    assert len(scheduled_jobs[1]) == 1
